from .transform import rotate90
from .stack import (stackmean, stacksum, stackvar, stackstderr, stackstd,
                    images_mean, images_sum)

__all__ = ['rotate90', 'stackmean', 'stacksum', 'stackvar', 'stackstderr',
           'stackstd', 'images_mean', 'images_sum']

# set version string using versioneer
from .._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np
from ..ext import fastccd
import time as ttime

import logging
logger = logging.getLogger(__name__)


def correct_images(images, dark=None, flat=None, gain=(1, 4, 8)):
    """Subtract backgrond and gain correct images

    This routine subtrtacts the backgrond and corrects the images
    for the multigain FastCCD ADC.

    Parameters
    ----------
    in : array_like
        Input array of images to correct of shape (N, y, x)  where N is the
        number of images and x and y are the image size.
    dark : array_like, optional
        Input array of dark images. This should be of shape (3, y, x).
        dark[0] is the gain 8 (most sensitive setting) dark image with
        dark[2] being the gain 1 (least sensitive) dark image.
    flat : array_like, optional
        Input array for the flatfield correction. This should be of shape
        (y, x)
    gain : tuple, optional
        These are the gain multiplication factors for the three different
        gain settings

    Returns
    -------
    array_like
        Array of corrected images of shape (N, y, x)

    """

    t = ttime.time()

    logger.info("Correcting image stack of shape %s", images.shape)

    if dark is None:
        dark = np.zeros(images.shape[-2:], dtype=np.float32)
        dark = np.array((dark, dark, dark))
        logger.info("Not correcting for darkfield. No input.")
    if flat is None:
        flat = np.ones(images.shape[-2:], dtype=np.float32)
        logger.info("Not correcting for flatfield. No input.")
    else:
        flat = np.asarray(flat, dtype=np.float32)

    data = fastccd.correct_images(images.astype(np.uint16),
                                  dark, flat, gain)
    t = ttime.time() - t

    logger.info("Corrected image stack in %.3f seconds", t)

    return data



from .images import correct_images
from .phocount import photon_count

__all__ = ['correct_images', 'photon_count']

# set version string using versioneer
from .._version import get_versions
__version__ = get_versions()['version']
del get_versions



from .animation import image_stack_to_movie, show_image_stack
from .nbviewer import notebook_to_nbviewer

# set version string using versioneer
from .._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = ['image_stack_to_movie', 'show_image_stack',
           'notebook_to_nbviewer']



from ..ext import phocount as ph


def photon_count(data, thresh, mean_filter, std_filter, nsum=3, nan=False):
    """Do single photon counting on CCD image

    This routine does single photon counting by cluster analysis. The image
    is searched for bright pixels within a threshold and then the energy
    deposited by each photon is calculated.

    Parameters
    ----------
    data : array_like
        Stack of CCD images. This array should be of shape (N, y, x) where
        N is the number of images
    thresh : tuple
        Threshold to use for identifying photons. This should be a tuple of
        (min, max)
    mean_filter : tuple
        Filter only the values of the mean which are within the limits of
        the tuple of the form (min, max)
    std_filter : tuple
        Filter only the values of the standard deviation  which are within
        the limits of the tuple of the form (min, max)
    nsum : int
        The number of pixels to use to calculate the energy deposited by the
        photon. This should be 0 < nsum <= 9.
    nan : bool
        If true, replace empty pixels with ``np.nan``

    Returns
    -------
    tuple
        Two arrays are returned. The first is an array of size (N, y, x)
        where the elements are the integrated energy calculated for each
        photon hit. The second array is the standard deviation for the
        integrated intensity on each photon hit.
    """
    return ph.count(data, thresh, mean_filter, std_filter, nsum, nan)



# Now import useful functions

from .utils import (get_fastccd_images, get_fastccd_timestamps)  # noqa F401
from .plotting import make_panel_plot  # noqa F401

# set version string using versioneer
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import numpy as np
from ..ext import image as extimage

import logging
logger = logging.getLogger(__name__)


def stackmean(array):
    """Cacluate the mean of a stack

    This function calculates the mean of a stack of images (or any array).
    It ignores values that are np.NAN and does not include them in the mean
    calculation. It assumes an array of shape (.. i, j, x, y) where x and y
    are the size of the returned array (x, y).

    Parameters
    ----------
    array : array_like
        Input array of at least 3 dimensions.

    Returns
    -------
    array
        2D Array of mean of stack.
    """
    X, Y = extimage.stackprocess(array, 1)
    return X


def stacksum(array, norm=True):
    """Cacluate the sum of a stack

    This function calculates the sum of a stack of images (or any array).
    It ignores values that are np.NAN and does not include them in the sum
    calculation. It assumes an array of shape (.. i, j, x, y) where x and y
    are the size of the returned array (x, y).

    The output sum is corrected for elements where NaNs are encountered if
    norm is set to True.  The values are renormalized to a sum which would
    have occured if all elements had been polulated. If no values are present
    in a stack, then NaN is returned.

    If norm is false, the actual sum is returned.

    size.

    Parameters
    ----------
    array : array_like
        Input array of at least 3 dimensions.

    Returns
    -------
    tuple
        tuple of 2 arrays of the sum and number of points in the sum
    """
    X, Y = extimage.stackprocess(array, 0)

    if norm:
        # Set zero values to NaN
        _Y = Y.astype(np.float32)
        _Y[Y == 0] = np.nan

        total_elements = array.size / (array.shape[-1] * array.shape[-2])

        X = X * (total_elements / _Y)

    return X, Y


def stackstd(array):
    """Cacluate the standard deviation of a stack

    This function calculates the standard deviation of a stack of images
    It ignores values that are np.NAN and does not include them in the sum
    calculation. It assumes an array of shape (.. i, j, x, y) where x and y
    are the size of the returned array (x, y).

    Parameters
    ----------
    array : array_like
        Input array of at least 3 dimensions.

    Returns
    -------
    tuple
        tuple of 2 arrays of the standard deviation and number of points
        in the calculation
    """
    X, Y = extimage.stackprocess(array, 3)
    return X, Y


def stackvar(array):
    """Cacluate the varience of a stack

    This function calculates the variance of a stack of images (or any array).
    It ignores values that are np.NAN and does not include them in the
    calculation. It assumes an array of shape (.. i, j, x, y) where x and y
    are the size of the returned array (x, y).

    Parameters
    ----------
    array : array_like
        Input array of at least 3 dimensions.

    Returns
    -------
    tuple
        tuple of 2 arrays of the varience and number of points in the
        calculation
    """
    X, Y = extimage.stackprocess(array, 2)
    return X, Y


def stackstderr(array):
    """Cacluate the standard error of a stack

    This function calculates the standard error of a stack of images
    (or any array).  It ignores values that are np.NAN and does not include
    them in the calculation. It assumes an array of shape (.. i, j, x, y)
    where x and y are the size of the returned array (x, y).

    Parameters
    ----------
    array : array_like
        Input array of at least 3 dimensions.

    Returns
    -------
    tuple
        tuple of 2 arrays of the standard error and number of points in the
        calculation
    """
    X, Y = extimage.stackprocess(array, 4)
    return X, Y


def images_mean(images):
    """Cacluate the mean ccd counts per event

    This function calculates the mean of ccd counts for each event. The
    input is a "slicerator" object returned by get_fastccd_images.

    Parameters
    ----------
    slicerator object : generator returning pims images
        This is the output of get_fastccd_images

    Returns
    -------
    array: 1D numpy array
    """
    return np.array([np.nanmean(stackmean(image)) for image in images])


def images_sum(images):
    """Cacluate the total ccd counts per event

    This function calculates the sum of ccd counts for each event. The
    input is a "slicerator" object returned by get_fastccd_images.

    Parameters
    ----------
    slicerator object : generator returning pims images
        This is the output of get_fastccd_images

    Returns
    -------
    array: 1D numpy array
    """
    return np.array([np.nansum(stackmean(image)) for image in images])



import numpy as np
import time as ttime
from pims import pipeline

from .fastccd import correct_images
from .image import rotate90, stackmean
from .settings import detectors
from databroker.assets.handlers import AreaDetectorHDF5TimestampHandler

import logging
logger = logging.getLogger(__name__)


def get_fastccd_images(light_header, dark_headers=None,
                       flat=None, gain=(1, 4, 8), tag=None, roi=None):
    """Retreive and correct FastCCD Images from associated headers

    Retrieve FastCCD Images from databroker and correct for:

    -   Bad Pixels (converted to ``np.nan``)
    -   Backgorund.
    -   Multigain bits.
    -   Flatfield correction.
    -   Rotation (returned images are rotated 90 deg cw)

    Parameters
    ----------
    light_header : databorker header
        This header defines the images to convert

    dark_headers : tuple of 3 databroker headers , optional
        These headers are the dark images. The tuple should be formed
        from the dark image sets for the Gain 8, Gain 2 and Gain 1
        (most sensitive to least sensitive) settings. If a set is not
        avaliable then ``None`` can be entered.

    flat : array_like
        Array to use for the flatfield correction. This should be a 2D
        array sized as the last two dimensions of the image stack.

    gain : tuple
        Gain multipliers for the 3 gain settings (most sensitive to
        least sensitive)

    tag : string
        Data tag used to retrieve images. Used in the call to
        ``databroker.get_images()``. If `None`, use the defualt from
        the settings.

    roi : tuple
        coordinates of the upper-left corner and width and height of
        the ROI: e.g., (x, y, w, h)

    Returns
    -------
        A corrected pims.pipeline of the data

    """

    if tag is None:
        tag = detectors['fccd']

    # Now lets sort out the ROI
    if roi is not None:
        roi = list(roi)
        # Convert ROI to start:stop from start:size
        roi[2] = roi[0] + roi[2]
        roi[3] = roi[1] + roi[3]
        logger.info("Computing with ROI of %s", str(roi))

    if dark_headers is None:
        bgnd = None
        logger.warning("Processing without dark images")
    else:
        if dark_headers[0] is None:
            raise NotImplementedError("Use of header metadata to find dark"
                                      " images is not implemented yet.")

        # Read the images for the dark headers
        t = ttime.time()

        dark = []
        for i, d in enumerate(dark_headers):
            if d is not None:
                # Get the images

                bgnd_events = _get_images(d, tag, roi)

                # We assume that all images are for the background
                # TODO : Perhaps we can loop over the generator
                # If we want to do something lazy

                tt = ttime.time()
                b = get_images_to_3D(bgnd_events, dtype=np.uint16)
                logger.info("Image conversion took %.3f seconds",
                            ttime.time() - tt)

                b = correct_images(b, gain=(1, 1, 1))
                tt = ttime.time()
                b = stackmean(b)
                logger.info("Mean of image stack took %.3f seconds",
                            ttime.time() - tt)

            else:
                if (i == 0):
                    logger.warning("Missing dark image"
                                   " for gain setting 8")
                elif (i == 1):
                    logger.warning("Missing dark image"
                                   " for gain setting 2")
                elif (i == 2):
                    logger.warning("Missing dark image"
                                   " for gain setting 1")

            dark.append(b)

        bgnd = np.array(dark)

        logger.info("Computed dark images in %.3f seconds", ttime.time() - t)

    events = _get_images(light_header, tag, roi)

    # Ok, so lets return a pims pipeline which does the image conversion

    # Crop Flatfield image
    if flat is not None and roi is not None:
        flat = _crop(flat, roi)

    return _correct_fccd_images(events, bgnd, flat, gain)


def get_images_to_4D(images, dtype=None):
    """Convert image stack to 4D numpy array

    This function converts an image stack from
    :func: get_images() into a 4D numpy ndarray of a given datatype.
    This is useful to just get a simple array from detector data

    Parameters
    ----------
    images : the result of get_images()
    dtype : the datatype to use for the conversion

    Example
    -------
    >>> header = DataBroker[-1]
    >>> images = get_images(header, "my_detector')
    >>> a = get_images_to_4D(images, dtype=np.float32)

    """
    im = np.array([np.asarray(im, dtype=dtype) for im in images],
                  dtype=dtype)
    return im


def get_images_to_3D(images, dtype=None):
    """Convert image stack to 3D numpy array

    This function converts an image stack from
    :func: get_images() into a 3D numpy ndarray of a given datatype.
    This is useful to just get a simple array from detector data

    Parameters
    ----------
    images : the result of get_images()
    dtype : the datatype to use for the conversion

    Example
    -------
    >>> header = DataBroker[-1]
    >>> images = get_images(header, "my_detector')
    >>> a = get_images_to_3D(images, dtype=np.float32)

    """
    im = np.vstack([np.asarray(im, dtype=dtype) for im in images])
    return im


def _get_images(header, tag, roi=None):
    t = ttime.time()
    if isinstance(header, (list, tuple)):
        # assumes all headers are coming from the same db
        db = header[0].db
        for h in header:
            if h.db is not db:
                raise ValueError("All headers need to come from the same "
                                 "Broker instance.")
        images = db.get_images(header, tag)
    else:
        images = header.db.get_images(header, tag)
    t = ttime.time() - t
    logger.info("Took %.3f seconds to read data using get_images", t)

    if roi is not None:
        images = _crop_images(images, roi)

    return images


@pipeline
def _correct_fccd_images(image, bgnd, flat, gain):
    image = correct_images(image, bgnd, flat, gain)
    image = rotate90(image, 'cw')
    return image


@pipeline
def _crop_images(image, roi):
    return _crop(image, roi)


def _crop(image, roi):
    image_shape = image.shape
    # Assuming ROI is specified in the "rotated" (correct) orientation
    roi = [image_shape[-2]-roi[3], roi[0], image_shape[-1]-roi[1], roi[2]]
    return image.T[roi[1]:roi[3], roi[0]:roi[2]].T


def get_fastccd_timestamps(header, tag='fccd_image'):
    """Return the FastCCD timestamps from the Areadetector Data File

    Return a list of numpy arrays of the timestamps for the images as
    recorded in the datafile.

    Parameters
    ----------
    header : databorker header
        This header defines the run
    tag : string
        This is the tag or name of the fastccd.

    Returns
    -------
        list of arrays of the timestamps

    """
    with header.db.reg.handler_context(
            {'AD_HDF5': AreaDetectorHDF5TimestampHandler}):
        timestamps = list(header.data(tag))

    return timestamps


def calculate_flatfield(image, limits=(0.6, 1.4)):
    """Calculate a flatfield from fluo data

    This routine calculates the flatfield correction from fluorescence data
    The image is thresholded by limits from the median value of the image.
    The flatfield is then constructed from the mean of the image divided by
    the masked (by NaN) image resulting in a true flatfield correction

    Parameters
    ----------
    image : array_like
        This is the 2D image to convert to a flatfield correction.
    limits : tuple
        Pixels outwith the median value multiplied by these limits will be
        excluded by setting to NaN.

    Returns
    -------
        Array of flatfield correction.

    """

    flat = image
    limits = np.array(limits)
    limits = np.nanmedian(image) * limits

    flat[flat < limits[0]] = np.nan
    flat[flat > limits[1]] = np.nan
    flat = np.nanmean(flat) / flat
    flat = np.rot90(flat)

    return flat



def get_fastccd_flatfield(light, dark, flat=None, limits=(0.6, 1.4), half_interval=False):
    """Calculate a flatfield from two headers 

    This routine calculates the flatfield using the
    :func:calculate_flatfield() function after obtaining the images from
    the headers.

    Parameters
    ----------
    light : databroker header
        The header containing the light images
    dark : databroker header(s)
        The header(s) from the run containin the dark images. See get_fastccd_images for details
    flat : flatfield image (optional)
        The array to be used for the initial flatfield
    limits : tuple limits used for returning corrected pixel flatfield
        The tuple setting lower and upper bound. np.nan returned value is outside bounds
    half_interval : boolean or tuple to perform calculation for only half of the FastCCD
        Default is False. If True, then the hard-code portion is retained.  Customize image 
        manipulation using a tuple of length 2 for (row_start, row_stop).


    Returns
    -------
    array_like
        Flatfield correction.  The correction is orientated as "raw data" not final data generated by get_fastccd_images().
    """
    images = get_images_to_3D(get_fastccd_images(light, dark, flat))
    images = stackmean(images)
    if half_interval:
        if isinstance(half_interval, bool):
            row_start, row_stop = (7, 486) #hard coded for the broken half of the fccd
        else:
            row_start, row_stop = half_interval
            print(row_start, row_stop)
        images[:, row_start:row_stop] = np.nan
    flat = calculate_flatfield(images, limits)
    removed = np.sum(np.isnan(flat))
    if removed != 0:
        logger.warning("Flatfield correction removed %d pixels (%.2f %%)" %
                       (removed, removed * 100 / flat.size))
    return flat


def fccd_mask():
    """Return the initial flatfield mask for the FastCCD

    Returns
    -------
    np.array of flatfield

    """
    flat = np.ones((960, 960))
    flat[120:250, 0:480] = np.nan
    flat[:, 476:484] = np.nan
    flat = np.rot90(flat)

    return flat



from __future__ import (absolute_import, division, print_function)
import setuptools
from distutils.core import setup, Extension
import numpy as np
import versioneer

with open('requirements.txt') as f:
    requirements = f.read().split()

with open('requirements-extras.txt') as f:
    extras_require = {'complete': f.read().split()}

fastccd = Extension('fastccd',
                    sources=['src/fastccdmodule.c',
                             'src/fastccd.c'],
                    extra_compile_args=['-fopenmp'],
                    extra_link_args=['-lgomp'])

image = Extension('image',
                  sources=['src/imagemodule.c',
                           'src/image.c'],
                  extra_compile_args=['-fopenmp'],
                  extra_link_args=['-lgomp'])

phocount = Extension('phocount',
                     sources=['src/phocountmodule.c',
                              'src/phocount.c'],
                     extra_compile_args=['-fopenmp'],
                     extra_link_args=['-lgomp'])
setup(
    name='csxtools',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Brookhaven National Laboratory',
    packages=setuptools.find_packages(exclude=['src', 'tests']),
    ext_package='csxtools.ext',
    include_dirs=[np.get_include()],
    ext_modules=[fastccd, image, phocount],
    tests_require=['pytest'],
    install_requires=requirements,
    extras_require=extras_require,
    url='http://github.com/NSLS-II_CSX/csxtools',
    keywords='Xray Analysis',
    license='BSD'
)



import numpy as np
from matplotlib import pyplot as plt

golden_mean = (np.sqrt(5)-1.0)/2.0


def make_panel_plot(n, fig=None,
                    xlmargin=0.15, ytmargin=0.10,
                    xrmargin=0.05, ybmargin=0.10):
    """Make a multi panel plot using matplotlib

    This function, makes a typical panel plot and returns a list
    of the axes objects for plotting later.

    Parameters
    ----------
    n : int
       Number of panels
    fig : figure object, optional
       Figure object to use (If None creates new figure)
    xmargin : float, optional
       Margin at x-axis
    ymargin : float, optional
       Margin at y-axis

    Returns
    -------
    tuple
        tuple of matplotlib axes objects for the panel plot

    """

    if fig is None:
        fig = plt.figure(figsize=[6, 6 * golden_mean * n])

    xsize = (1. - (xlmargin + xrmargin))
    ysize = (1. - (ybmargin + ytmargin)) / n

    pos = np.array([xlmargin, ybmargin, xsize, ysize])

    allax = []
    for x in range(n):
        ax = fig.add_axes(pos + np.array([0, ysize * x, 0, 0]))
        if x > 0:
            # Remove ticklabels
            ax.xaxis.set_ticklabels("")
        allax.append(ax)

    return allax



from itertools import chain
import uuid
import time as ttime
from metadatastore.commands import get_events_generator
from databroker.databroker import fill_event
from databroker.pivot import pivot_timeseries, zip_events, reset_time


def correct_events(evs, data_key, dark_images, drop_raw=False):
    out_data_key = data_key + '_corrected'
    ev0 = next(evs)
    new_desc = dict(ev0['descriptor'])
    new_desc['data_keys'][out_data_key] = dict(new_desc['data_keys'][data_key])
    new_desc['data_keys'][out_data_key]['source'] = 'subtract_background'
    new_desc['uid'] = uuid.uuid4()
    if drop_raw:
        new_desc['data_keys'].pop(data_key)
    for ev in chain((ev0, ), evs):
        new_ev = {'uid': str(uuid.uuid4()),
                  'time': ttime.time(),
                  'descriptor': new_desc,
                  'seq_no': ev['seq_no'],
                  'data': dict(ev['data']),
                  'timestamps': dict(ev['timestamps'])}
        corr, gain_img = subtract_background(ev['data'][data_key], dark_images) # noqa F821 TODO
        new_ev['data'][out_data_key] = corr
        new_ev['timestamps'][out_data_key] = ttime.time()
        if drop_raw:
            new_ev['data'].pop(data_key)
            new_ev['timestamps'].pop(data_key)
        yield new_ev


def clean_images(header, pivot_key, timesource_key, dark_images=None, static_keys=None):
    if static_keys is None:
        static_keys = ['sx', 'sy', 'temp_a', 'temp_b', 'sz']
    # sort out which descriptor has the key we want to pivot on
    pv_desc = [d for d in header['descriptors'] if pivot_key in d['data_keys']][0]
    # sort out which descriptor has the key that we want to zip with to get time stamps
    ts_desc = [d for d in header['descriptors'] if timesource_key in d['data_keys']][0]

    ts_events = get_events_generator(ts_desc)
    pv_events = get_events_generator(pv_desc)
    # fill_event works in place, sillyness to work around that
    pv_events = ((ev, fill_event(ev))[0] for ev in pv_events)
    pivoted_events = pivot_timeseries(pv_events, [pivot_key], static_keys)

    if dark_images:
        pivoted_events = correct_events(pivoted_events, pivot_key, dark_images)
    merged_events = zip_events(pivoted_events, ts_events)
    out_ev = reset_time(merged_events, timesource_key)
    yield from out_ev


def extract_darkfield(header, dark_key):
    cam_desc = [d for d in header['descriptors'] if dark_key in d['data_keys']][0]
    events = get_events_generator(cam_desc)
    events = list(((ev, fill_event(ev))[0] for ev in events))
    event = events[0]
    ims = (event['data'][dark_key] << 2) >> 2
    return ims.mean(axis=0)



from ..ext import image as extimage


def rotate90(a, sense='ccw'):
    """Rotate a stack of images by 90 degrees

    This routine rotates a stack of images by 90. The rotation is performed
    on the last two axes. i.e. For a stack of images of shape (N, y, x)
    N rotations of the image of size (y, x) are performed.

    Parameters
    ----------
        a : array_like
            Input array to be rotated. This should be of shape (N, y, x).
        sense : string
            'cw' to rotate clockwise, 'ccw' to rotate anitclockwise

    Returns
    -------
        array
            Rotated stack of images of shape (N, x, y)

    """

    if sense == 'ccw':
        sense = 1
    elif sense == 'cw':
        sense = 0
    else:
        raise ValueError("sense must be 'cw' or 'ccw'")

    return extimage.rotate90(a, sense)
